@@ -1,5 +1,6 @@
 from functools import wraps
 from importlib.util import find_spec
+from rdkit.Chem import MolToSmiles
 
 def requires(module):
     def inner(func):
@@ -31,4 +32,14 @@ def tooltip_formatter(s, subset, fmt, style):
     for k, v in s[subset].to_dict().items():
         v = f'<span style="{style[k](v)}">{v}</span>' if style.get(k) else v
         items.append(fmt.format(key=k, value=v))
-    return "<br>".join(items)
\ No newline at end of file
+    return "<br>".join(items)
+
+def mol_to_smiles(mol):
+    """Returns a SMILES from an RDKit molecule, or None if not an RDKit mol"""
+    return MolToSmiles(mol) if mol else None
+
+def mol_to_record(mol, mol_col="mol"):
+    """Function to create a dict of data from an RDKit molecule"""
+    return {"SMILES": MolToSmiles(mol),
+            **mol.GetPropsAsDict(),
+            mol_col: mol} if mol else {}
\ No newline at end of file
